/**
* @file  file_device_zip.cpp
* @brief Implementierung der Klasse FileDeviceZip.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_device_zip.h"
#include <sys/types.h>
#include <sys/stat.h>

#include <list>
#include <cstdio>
#include <memory>

#include "common/device_update.h"
#include "common/utils.h"

#include <boost/format.hpp>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#if _MSVC_LANG >= 201703L
#include <filesystem>
namespace fs = std::filesystem;
#endif

#include "common/md5.h"

FileDeviceZip::FileDeviceZip()
{
    // Initialisieren
    resetData();

    // Auch den Error mitresetten.
    resetError();
}

FileDeviceZip::~FileDeviceZip()
{
}

void FileDeviceZip::resetData()
{
    // Basis Reset
    FileDevice::resetData();
}

void FileDeviceZip::resetError()
{
    // Basis Reset
    FileDevice::resetError();
    // Fehlertext loeschen
    m_errorExtended.clear();
}

void FileDeviceZip::setRecentZipError(const std::string &filePath, int lineNumber)
{
    std::string text;
#if defined(_MSC_VER) || defined(__MINGW32__)
    // Fehlertext ermitteln.
    text.resize(255);
    FormatZipMessage( ZR_RECENT, &text[0], static_cast<unsigned int>( text.size() ) );
#endif
    setErrorText( text );
    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % filePath % lineNumber % text ) );
}

void FileDeviceZip::setErrorTextExtended( const std::string &text )
{
    m_errorExtended = text;
}

void FileDeviceZip::appendErrorTextExtended( const std::string &text )
{
    m_errorExtended.append( text );
}

std::string FileDeviceZip::errorTextExtended()
{
    return m_errorExtended;
}

int FileDeviceZip::CreateZipFromDirectory(char *path, char *fileZIP, char *defaultDFL, const std::string &updateDBPath )
{
    int loaded = 0;

#if defined(_MSC_VER) || defined(__MINGW32__)
    char szZipFile[260], szFile[260], szIniFile[260];
    char szSearchFile[260];
    intptr_t fp;
    struct _finddata_t finddata;
    std::string version;
    std::list<std::string> versions;
    std::string timestamp;
    std::string comment;
    int fileCount = 0;
    std::unique_ptr<char, void (*)( void *)> md5( nullptr, &free );

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Speicher initialisieren
    memset(szZipFile, 0, sizeof(szZipFile));
    memset(szFile, 0, sizeof(szFile));
    memset(szIniFile, 0, sizeof(szIniFile));
    memset(szSearchFile, 0, sizeof(szSearchFile));

    // Je nach Angabe Basispfad mit berücksichtigen
    if (fileZIP[1] == ':') {

        strcpy(szZipFile, fileZIP);
    } else {

        sprintf(szZipFile, "%s\\%s", path, fileZIP);
    }

    sprintf(szIniFile, "%s\\%s", path, "SWI.ini");
    ::remove(szIniFile); // Vorhandene Initialisierungsdatei loeschen

    sprintf(szSearchFile, "%s\\*.hex", path);
    // Passende Dateien fuer Muster?
    if ((fp = _findfirst(szSearchFile, &finddata)) != -1) {

        // Vorhandenes Archiv loeschen
        ::remove(szZipFile); 
        // Zipdatei erzeugen
        HZIP hZip = NULL;
        if ((hZip = CreateZip(szZipFile, 0)) == NULL) {

            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }

        // Builddatum, wenn nicht vorhanden, eintragen
        if (timestamp.size() != 19) {

            char tmp[20];
            time_t ltime;
            struct tm *ltm;

            time(&ltime);
            ltm = localtime(&ltime);
            strftime(tmp, sizeof(tmp), "%Y-%m-%d %H:%M:%S", ltm);
            timestamp = std::string( tmp );
        }

        // Alle Dateien durchlaufen.
        do {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s\\%s", path, finddata.name);

            // Umgebung zuruecksetzen
            resetData();
            // Fehlerobjekt zuruecksetzen
            resetError();

            // Dateidaten laden
            if (!load(szFile)) {

                // Eigenen Fehlertext setzen.
                setErrorText( boost::str( boost::format( "Fehler (%1%) beim Ladeversuch der Datei [ %2% ]" ) %  static_cast<int>( errorState() ) % szFile ) );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                goto error_label;
            }

            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            if (GetHeaderSoftwareInfo().size) {

                // MD5 der Datei in Initialisierungsdatei uebernehmen
                md5.reset( MD5File(szFile) );
                if ( !md5 ) {

                    // Datei kann nicht uebernommen werden.
                    // Eigenen Fehlertext setzen.
                    setErrorText( "Erstellen des MD5(1) schlug fehl." );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                    goto error_label;
                }

                m_iniFile.WriteString(finddata.name, "MD5", md5.get());

                // Version zur Datei in Initialisierungsdatei uebernehmen.
                GetHeaderVersionFirmware( version );
                if ( version.length() > 11 ) // Version nur als 00.00.00.00 Format zulassen, ggf. Endung ZK, PZE etc. wegschneiden.
                    version.resize( 11 );

                m_iniFile.WriteString(finddata.name, "Version", version.data());

                // Versionen fuer Archivini halten.
                // Pruefung ob diese Version bereits eingetragen ist.
                if ( std::find( versions.begin(), versions.end(), version ) == versions.end() )
                    versions.push_back( version );

                // Softwareinfo zur Datei in Initialisierungsdatei uebernehmen.
                std::string tmp;
                StructSoftwareInfo swInfo = GetHeaderSoftwareInfo();
                SoftwareInfoStructToString(tmp, swInfo);
                m_iniFile.WriteString(finddata.name, "SWI", tmp.data());

                // Datei zu Archiv hinzufuegen
                if (ZipAdd(hZip, finddata.name, szFile) != ZR_OK) {
                    // Fehlertext von Zip-API setzen.
                    setRecentZipError( __FILE__, __LINE__ );
                    goto error_label;
                }

                fileCount++;
#if _MSVC_LANG >= 201703L
                // Dateien auch als IFF-Format eintragen.
                if ( SoftwareVersion( version ) > "04.03.00.00" )
                {
                    fs::path iff_file_path( szFile );
                    iff_file_path.replace_extension( "iff" );

                    if ( fs::exists( iff_file_path ) )
                    {
                        // MD5 der Datei in Initialisierungsdatei uebernehmen
                        md5.reset( MD5File( iff_file_path.string().c_str() ) );
                        if ( !md5 )
                        {
                            // Datei kann nicht uebernommen werden.
                            // Eigenen Fehlertext setzen.
                            setErrorText( "Erstellen des MD5(1) schlug fehl." );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                            goto error_label;
                        }

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "MD5", md5.get() );

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "Version", version.data() );

                        m_iniFile.WriteString( iff_file_path.filename().string().c_str(), "SWI", tmp.data() );

                        // Datei zu Archiv hinzufuegen
                        if ( ZipAdd( hZip, iff_file_path.filename().string().c_str(), iff_file_path.string().c_str() ) != ZR_OK )
                        {
                            // Fehlertext von Zip-API setzen.
                            setRecentZipError( __FILE__, __LINE__ );
                            goto error_label;
                        }

                        fileCount++;
                    }
                }
#endif
            }

        } while (_findnext(fp, &finddata) != -1);

        // Textdaten fuer Firmware verfuegbar?
        if (defaultDFL != NULL) {

            // MD5 der Datei in Initialisierungsdatei uebernehmen
            md5.reset( MD5String( defaultDFL ) );
            if ( !md5 ) {

                // Datei kann nicht uebernommen werden.
                // Eigenen Fehlertext setzen.
                setErrorText( "Erstellen des MD5(1) schlug fehl." );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
                goto error_label;
            }

            m_iniFile.WriteString("Default.dfl", "MD5", md5.get() );

            // Datei mit Textdaten schreiben
            if (ZipAdd(hZip, "Default.dfl", defaultDFL, (unsigned int)strlen(defaultDFL)) != ZR_OK) {
                // Fehlertext von Zip-API setzen.
                setRecentZipError( __FILE__, __LINE__ );
                goto error_label;
            }
        }

        // Updatedatenbank zum Zip-Archiv hinzufügen
        if ( !updateDBPath.empty() )
        {
            constexpr char *dbFileName = "firmwareUpdate.db";

            // MD5 berechnen
            md5.reset( MD5File( updateDBPath.c_str() ) );
            if ( !md5 )
            {
                // Datei kann nicht uebernommen werden.
                // Eigenen Fehlertext setzen.
                setErrorText( "Erstellen des MD5(1) schlug fehl." );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                goto error_label;
            }
            m_iniFile.WriteString( dbFileName, "MD5", md5.get() );

            // Datei zu Archiv hinzufügen
            if ( ZipAdd( hZip, dbFileName, updateDBPath.c_str() ) != ZR_OK )
            {
                // Fehlertext von Zip-API setzen.
                setRecentZipError( __FILE__, __LINE__ );
                goto error_label;
            }
        }

        // Initialisierungsdatei schreiben
        m_iniFile.WriteIniFile(szIniFile);
        m_iniFile.CloseIniFile();
        // Datei zu Archiv hinzufuegen
        if (ZipAdd(hZip, "SWI.ini", szIniFile) != ZR_OK) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }

        // MD5 der Initialisierungsdatei als Name fuer Sicherheitsinitialisierungsdatei erstellen.
        md5.reset( MD5File( szIniFile ) );
        if ( !md5 )
        {
            // Datei kann nicht uebernommen werden.
            // Eigenen Fehlertext setzen.
            setErrorText( "Erstellen des MD5(2) schlug fehl." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            goto error_label;
        }
        // Aufgenommene Initialisierungsdatei loeschen
        ::remove(szIniFile);

        sprintf(szFile, "%s.ini", md5.get());

        sprintf(szIniFile, "%s\\%s", path, szFile);
        ::remove(szIniFile); // Vorhandene Initialisierungsdatei loeschen

        // Archivdaten eintragen
        m_iniFile.WriteInt("Archive", "FileDeviceVersion", sizeof(SDFFileDeviceHeader));
        m_iniFile.WriteString("Archive", "Timestamp", timestamp.data());
        if ( versions.size() > 1 )
        {
            int idx = 1;
            std::string key;
            versions.sort( std::greater<std::string>() ); // Liste der Versionen aufsteigend sortieren
            for ( std::string &ver : versions )
            {
                version.clear();
                key = boost::str( boost::format( "Version%1%" ) % idx++ );
                std::copy_n( ver.begin(), 11, std::back_inserter( version ) );

                m_iniFile.WriteString( "Archive", key.data(), version.data() );
            }
        }
        else
        {
            m_iniFile.WriteString("Archive", "Version", versions.begin()->data() );
        }

        m_iniFile.WriteInt("Archive", "Count", fileCount);

        // Initialisierungsdatei schreiben
        m_iniFile.WriteIniFile(szIniFile);
        m_iniFile.CloseIniFile();

        // Datei zu Archiv hinzufuegen
        if (ZipAdd(hZip, szFile, szIniFile) != ZR_OK) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            goto error_label;
        }
        // Aufgenommene Initialisierungsdatei loeschen
        ::remove(szIniFile);

        // Erstellen erfolgreich
        loaded = 1;
        // Kommentar des Archives setzen
        comment = boost::str( boost::format( "Company:\tDatafox GmbH\nContent:\tFiles for Datafox - devices.\nVersion:\t%1%\nTime stamp:\t%2%") % versions.begin()->data() % timestamp );
        ZipSetComment(hZip, comment.data());
error_label:
        _findclose(fp);
        // Zipdatei schliessen
        CloseZip(hZip);
        m_iniFile.CloseIniFile();
    } else {

        // Eigenen Fehlertext setzen.
        setErrorText( boost::str( boost::format( "Keine Hex-Dateien im Pfad [ %1% ] gefunden." ) %  path ) );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 

    }

    // Reset der Daten durchfuehren
    resetData();
#endif

    return loaded;
}

int FileDeviceZip::SearchInfosDir(std::string &filePath, const StructHardwareInfo &hwInfo, DeviceUpdate *deviceUpdate )
{
    if ( deviceUpdate == nullptr )
    {
        return 0;
    }

    int loaded = 0;
    char szFile[260]{};
    char tmp[260]{};
    std::string iniVersion;
    std::string errorExtended;

    bool blacklisted{};
    bool versionToOld{};
    bool versionUndefined{};

    std::function<int()> exitFunc = [&]()
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        m_iniFile.CloseIniFile();
#endif

    // Reset bei fehlerhaftem Laden
        if ( errorState() == Error_state::none && !loaded )
        {
            resetData();

            // Es wurde keine passende Datei gefunden.
            setErrorState( Error_state::no_file_match );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

            // Fehlertext der einzelnen Dateien uebernehmen.
            setErrorTextExtended( errorExtended );
        }
        else
        {
            // Fehlertext uebernehmen
            filePath += ":[" + errorText() + "]";
        }

        return loaded;
    };

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    std::string iniFilePath = filePath + "/SWI.ini";
#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( !m_iniFile.OpenIniFile( iniFilePath.c_str() ) )
#else
    if ( m_iniFile.LoadFile( iniFilePath.c_str()) < 0 )
#endif
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Laden der Softwareinfo fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc();
    }

    // Herstellung der Verbindung zur Datenbank für den Firmware-Hardware-Abgleich
    bool checkBoardRevision = true;
    if ( deviceUpdate != nullptr )
    {
        std::string errorMsg;
        checkBoardRevision = !deviceUpdate->open( errorMsg );
        if ( !errorMsg.empty() )
        {
            setErrorText( errorMsg );
        }
    }

#if defined(_MSC_VER) || defined(__MINGW32__)
    char szSearchFile[260];
    intptr_t fp;
    struct _finddata_t finddata;

    sprintf(szSearchFile, "%s\\*.hex", filePath.c_str());
    // Passende Dateien fuer Muster?
    if ((fp = _findfirst(szSearchFile, &finddata)) != -1) {

        // Alle Dateien durchlaufen.
        do {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s\\%s", filePath.c_str(), finddata.name);
#else
    DIR *d;
    struct stat st;
    struct dirent *de;

    if ((d = opendir(filePath.c_str())) != NULL) {

        // Alle Dateien durchlaufen
        while((de = readdir(d)) != NULL) {

            // Dateiname zusammenstellen
            sprintf(szFile, "%s/%s", filePath.c_str(), de->d_name);

            if (stat(szFile, &st) == -1) {
                break;
            }

            if (S_ISREG(st.st_mode)) {

                // Nur ".hex" gueltig.
                int pos = strlen(de->d_name) - 4;
                if (de->d_name[pos] != '.' || 
                    !(de->d_name[pos + 1] == 'h' || de->d_name[pos + 1] == 'H') ||
                    !(de->d_name[pos + 2] == 'e' || de->d_name[pos + 2] == 'E') ||
                    !(de->d_name[pos + 3] == 'x' || de->d_name[pos + 3] == 'X') ) {

                        continue;
                }

#endif	
                // Laden der Softwareinfo
#if defined(_MSC_VER) || defined(__MINGW32__)
                strcpy( tmp, m_iniFile.ReadString( finddata.name, "SWI", "" ) );
#else
                strcpy( tmp, m_iniFile.GetValue( de->d_name, "SWI", "" ) );
#endif
                if ( !strlen( tmp ) || tmp[0] != '#' )
                {
                    // Eigenen Fehlertext setzen.
                    setErrorText( "Lesen der Softwareinfo(1) fehlgeschlagen." );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return exitFunc();
                }

                // Die Softwareinfo in eine Struktur wandeln
                StructSoftwareInfo swInfo = SoftwareInfoStringToStruct( tmp );
                if ( !swInfo.size )
                {
                    // Eigenen Fehlertext setzen.
                    setErrorText( "Lesen der Softwareinfo(2) fehlgeschlagen." );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return exitFunc();
                }

                // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
#if defined(_MSC_VER) || defined(__MINGW32__)
                iniVersion = m_iniFile.ReadString( finddata.name, "Version", "" );
#else
                iniVersion = m_iniFile.GetValue( de->d_name, "Version", "" );
#endif

                // Die aktuelle Firmwaredatei wird ignoriert, wenn einer der folgenden Fälle eintritt:
                // (1) Version steht auf der Blacklist
                // (2) Version ist älter als die erlaubte Minimalversion (Bei der Ermittlung der Minimalversion wird die Whitelist berücksichtigt.)
                // (3) Es sollte eine minimale Version geben (checkBoardRevision == false), diese ist allerdings undefiniert
                // (4) Die Hardwareinfo passt nicht zur Softwareinfo
                //
                // Es genügt, wenn einer der Fälle eintritt.
                blacklisted = deviceUpdate->isVersionBlacklisted( iniVersion );
                versionToOld = deviceUpdate->minFwVersion() > SoftwarebuildVersion( iniVersion );
                versionUndefined = deviceUpdate->minFwVersion() == SoftwarebuildVersion();

                if ( blacklisted || versionToOld || (!checkBoardRevision && versionUndefined) ||
                     !CompareInfos( hwInfo, swInfo, checkBoardRevision ) )
                {
                    errorExtended.append( boost::str( boost::format( "%1%:[%2%]\n" ) % szFile % _deviceInfo.m_errorText ) );

                    // Naechste Datei pruefen
                    continue;
                }

#if defined(_MSC_VER) || defined(__MINGW32__)
                deviceUpdate->addFwVersion( iniVersion, finddata.name );
#else
                deviceUpdate->addFwVersion( iniVersion, de->d_name );
#endif

#if defined(_MSC_VER) || defined(__MINGW32__)
            } while (_findnext(fp, &finddata) != -1);
            _findclose(fp);
        }
#else
            }
        } // while
        closedir(d);
    }
#endif

    return exitFunc();
}

int FileDeviceZip::CheckZip(const std::string &filePath)
{
    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    std::string szFile;
    std::string szIniFile;
    std::string version;
    char timestamp[20];
    int index, fileDeviceVersion;
    ZIPENTRY ze;
    std::unique_ptr<char, void (*)( void * )> md5( nullptr, &free );

    // Speicher initialisieren
    memset(timestamp, 0, sizeof(timestamp));

    HZIP hZip = NULL;
    if ((hZip = OpenZip(filePath.c_str(), 0)) == NULL) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    szIniFile = Utils::File::uniqueTempFileName( "df_" );
    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile.c_str() ) ) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }
    // MD5 ueber die Initialisierungsdatei erstellen
    md5.reset( MD5File( szIniFile.c_str() ) );
    if ( !md5 ) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Dateiname zusammenstellen
    szFile = md5.get();
    szFile +=".ini";
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile.c_str(), false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    szFile = Utils::File::uniqueTempFileName( "df_" );
    // Entzippen der Sicherheitsinitialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szFile.c_str())) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Informationen der Sicherheitsini laden
    if (!m_iniFile.OpenIniFile(szFile.c_str())) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei ist gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Version der enthaltenen Geraetedateien pruefen
    fileDeviceVersion = m_iniFile.ReadInt("Archive", "FileDeviceVersion", 0);
    if (static_cast<size_t>( fileDeviceVersion ) > sizeof(SDFFileDeviceHeader)) {

        // Fehlertext zuweisen. Die Software ist nicht aktuell genug fuer diese Geraetedateien.
        setErrorState( Error_state::version );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto error_label;
    }

    // Anzahl der Dateien laden
    //fileCount = m_iniFile.ReadInt("Archive", "Count", 0);

    // Versionen der Dateien laden
    m_version.clear();
    int idx;
    for (idx = 0; idx < 11; ++idx)
    {
        std::string key;
        if (idx == 0)
        {
            key = "Version";
        }
        else
        {
            key = "Version" + std::to_string( idx );
        }

        version = m_iniFile.ReadString("Archive", key.c_str(), "");
        if (idx > 1 && version.size() == 0)
        {
            break;
        }

        if (version.size())
        {
            if (m_version.size())
                m_version += "; ";

            // Version halten
            m_version += version;
        }
    }

    // Zeitstempel des Archives laden
    strcpy(timestamp, m_iniFile.ReadString("Archive", "Timestamp", ""));
    if (!strlen(timestamp)) {
        // Zeitstempel anhand des Erstellungsdatums setzen
        struct stat st;
        if (stat(filePath.c_str(), &st) == 0) {

            struct tm *ptm = localtime(&st.st_ctime);

            sprintf(timestamp, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
                (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
                ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
        }
    }

    // Zeitstempel halten
    m_timestamp = timestamp;
    // Als geladen markieren
    loaded = 1;

error_label:
    // Schliessen der Initialisierungsdatei
    m_iniFile.CloseIniFile();
    // Dateien loeschen
    ::remove(szFile.c_str());
    ::remove(szIniFile.c_str());
    // Ziparchiv schliessen
    CloseZip(hZip);	
#endif
    return loaded;
}

int FileDeviceZip::UnzipDefaultDFL(char *fileZIP, char *fileDefaultDFL)
{
    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    std::string szFile;
    std::string szIniFile;
    std::string szFileDefaultDFL;
    char tmp[256];
    int index;
    ZIPENTRY ze;
    std::unique_ptr<char, void (*)( void *)> md5( nullptr, &free );

    // Speicher initialisieren
    memset(tmp, 0, sizeof(tmp));

    resetError();

    HZIP hZip = NULL;
    if ((hZip = OpenZip(fileZIP, 0)) == NULL) {

        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    szIniFile = Utils::File::uniqueTempFileName( "df_" );
    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile.c_str())) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // MD5 ueber die Initialisierungsdatei erstellen
    md5.reset( MD5File( szIniFile.c_str() ) );
    if ( !md5 ) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Dateiname der Sicherheitsini zusammenstellen
    szFile = md5.get();
    szFile += ".ini";
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile.c_str(), false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    if (!m_iniFile.OpenIniFile(szIniFile.c_str())) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // MD5 der Textdatedatei pruefen
    // Version der Dateien laden
    strcpy(tmp, m_iniFile.ReadString("Default.dfl", "MD5", ""));
    if (!strlen(tmp)) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Ermittlung des MD5 zur Default.dfl fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label;
    }

    // Suchen der Textdatendatei Default.dfl
    if (ZR_OK != FindZipItem(hZip, "Default.dfl", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // Dateiname zusammenstellen
    szFileDefaultDFL = Utils::File::uniqueTempFileName( "df_" );
    // Entzippen der Textdatendatei
    if (ZR_OK != UnzipItem(hZip, index, szFileDefaultDFL.c_str())) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

    // MD5 ueber die Textdatendatei erstellen
    md5.reset( MD5File( szFileDefaultDFL.c_str() ) );
    if ( !md5 ) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5 gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    if (strcmp(tmp, md5.get()) != 0) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Pruefung des MD5 der Default.dfl gescheitert." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto error_label; 
    }

    // Kopie der Datei an gewuenschte lokation vornehmen
    if (ZR_OK != UnzipItem(hZip, index, fileDefaultDFL)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto error_label;
    }

error_label:
    // Inidatei schliessen
    m_iniFile.CloseIniFile();
    // Dateien loeschen
    ::remove(szIniFile.c_str());
    ::remove(szFileDefaultDFL.c_str());
    // Ziparchiv schliessen
    CloseZip(hZip);	
#endif
    return loaded;
}

bool FileDeviceZip::unzipUpdateDB( const std::string &fileZip, const std::string &expectedMD5, std::string &dbPath )
{
#if defined(_MSC_VER) || defined(__MINGW32__)

    HZIP hZip = OpenZip( fileZip.c_str(), 0 );
    if ( hZip == nullptr )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return false;

    }

    std::function<bool( bool )> exitFunc = [&]( bool result )
    {
        CloseZip( hZip );
        return result;
    };

    ZIPENTRY ze{};
    int index{};
    const std::string dbFileName = "firmwareUpdate.db";

    // Suchen der Update-Datenbank
    if ( FindZipItem( hZip, dbFileName.c_str(), false, &index, &ze ) != ZR_OK )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return exitFunc( false );
    }

    // Entpacken der Datenbank aus dem Archiv
    int loop = 3;
    do
    {
        dbPath = Utils::File::uniqueTempFileName( "df_", ".db" ); // Pfad zur entpackten Datenbank aus dem Update-Archiv/-Ordner

        if ( !Utils::File::exists( dbPath ) )
        {
            break;
        }

    } while ( --loop );

    if ( !loop )
    {
        std::string errorText = std::string( "Error creating temporary database file. " ) + strerror( errno );
        DF_ASSERT( false, errorText.c_str() );
        return exitFunc( false );
    }

    if ( UnzipItem( hZip, index, dbPath.c_str() ) != ZR_OK )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return exitFunc( false );
    }

    // Prüfen des MD5, damit mit nur gültige, nicht manipulierte Datenbanken verwendet werden.
    return exitFunc( (expectedMD5.compare( std::unique_ptr<char, void (*)( void * )>( MD5File( dbPath.c_str() ), &free ).get() ) == 0) );

#else

    return false;

#endif
}

int FileDeviceZip::SearchInfosZip( std::string &filePath, const StructHardwareInfo &hwInfo, DeviceUpdate *deviceUpdate, std::function<void ( const char * )> ptrace )
{
    if ( deviceUpdate == nullptr )
    {
        return 0;
    }

    auto trace = [ptrace]( const char *msg )
    {
        if ( ptrace != nullptr )
        {
            ptrace( msg );
        }
    };

    int loaded = 0;
#if defined(_MSC_VER) || defined(__MINGW32__)
    std::string szFile;
    std::string szIniFile;
    std::string version;
    char timestamp[20], tmp[256];
    std::string errorExtended;
    int i, index, fileDeviceVersion;
    ZIPENTRY zeZIP, ze;
    std::unique_ptr<char, void(*)( void * )> md5( nullptr, &free );
    std::string iniVersion;

    bool blacklisted{};
    bool versionToOld{};
    bool versionUndefined{};
    bool checkBoardRevision = true;

    // Speicher initialisieren
    memset(timestamp, 0, sizeof(timestamp));
    memset(tmp, 0, sizeof(tmp));

    // Fehler zuruecksetzen
    resetError();

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Oeffnen des Zip-Archives ]\n" ) % __FILE__ % __LINE__ % __func__ ).c_str() );

    HZIP hZip = NULL;
    if ((hZip = OpenZip(filePath.c_str(), 0)) == NULL) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );

        // Fehler anfuegen
        filePath += ":[" + errorText() + "]";
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if (ZR_OK != FindZipItem(hZip, "SWI.ini", false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );

        goto error_label;
    }

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Finden von Datei 'SWI.ini', index = %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % index ).c_str() );

    // Dateiname zusammenstellen  
    szIniFile = Utils::File::uniqueTempFileName( "df_" );

    // Entzippen der Initialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szIniFile.c_str())) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Entzippen von 'SWI.ini' als %4%, TMP_MAX = %5% ]\n" ) % __FILE__ % __LINE__ % __func__ % szIniFile % TMP_MAX ).c_str() );

    // MD5 ueber die Initialisierungsdatei erstellen
    md5.reset( MD5File( szIniFile.c_str() ) );
    if ( !md5 ) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label; 
    }

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Erstellung des MD5 von extrahierter SWI.ini = %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % md5.get() ).c_str() );

    // Dateiname zusammenstellen
    szFile = md5.get();
    szFile += ".ini";
    // Suchen der Initialisierungsdatei der Sicherheit
    if (ZR_OK != FindZipItem(hZip, szFile.c_str(), false, &index, &ze)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Finden der Datei %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % szFile ).c_str() );

    // Dateiname zusammenstellen
    szFile = Utils::File::uniqueTempFileName( "df_" );
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Erstellung des Dateinamens %4%, index = %5%, hZip = %6% ]\n" ) % __FILE__ % __LINE__ % __func__ % szFile % index % hZip).c_str() );
    // Entzippen der Sicherheitsinitialisierungsdatei
    if (ZR_OK != UnzipItem(hZip, index, szFile.c_str())) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Entzippen von '%4%.ini' als %5% ]\n" ) % __FILE__ % __LINE__ % __func__ % md5.get() % szFile ).c_str() );

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Oeffnen der Inidatei(1) %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % szFile ).c_str() );

    // Informationen der Sicherheitsini laden
    if (!m_iniFile.OpenIniFile(szFile.c_str())) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }

    // Version der enthaltenen Geraetedateien pruefen
    fileDeviceVersion = m_iniFile.ReadInt("Archive", "FileDeviceVersion", 0);

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ FileDeviceVersion aus Inidatei(1) ist %4% ]\n" ) % __FILE__ %__LINE__ %__func__ % fileDeviceVersion ).c_str() );

    if (static_cast<unsigned>( fileDeviceVersion ) > sizeof(SDFFileDeviceHeader)) {

        // Fehlertext zuweisen. Die Software ist nicht aktuell genug fuer diese Geraetedateien.
        setErrorState( Error_state::version );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }

    // Anzahl der Dateien laden
    //fileCount = m_iniFile.ReadInt("Archive", "Count", 0);
    // Version der Dateien laden
    version = m_iniFile.ReadString("Archive", "Version", "");
    // Zeitstempel des Archives laden
    strcpy(timestamp, m_iniFile.ReadString("Archive", "Timestamp", ""));
    if (!strlen(timestamp)) {
        // Zeitstempel anhand des Erstellungsdatums setzen
        struct stat st;
        if (stat(filePath.c_str(), &st) == 0) {

            struct tm *ptm = localtime(&st.st_ctime);

            sprintf(timestamp, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
                (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
                ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
        }
    }

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Schliessen der Inidatei(1) %4% ]\n" ) % __FILE__ %__LINE__ %__func__ % szFile ).c_str() );

    // Schliessen der Initialisierungsdatei
    m_iniFile.CloseIniFile();
    // Datei loeschen
    ::remove(szFile.c_str());

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Infos zum Zip-Archive ermitteln ]\n" ) % __FILE__ % __LINE__ % __func__ ).c_str() );

    // Informationen zum Archiv ermitteln
    if (ZR_OK != GetZipItem(hZip, -1, &zeZIP)) {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }

    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Oeffnen der Inidatei(2) %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % szIniFile ).c_str() );

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    if (!m_iniFile.OpenIniFile(szIniFile.c_str())) {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
        goto error_label;
    }

    // Herstellung der Verbindung zur Datenbank für den Firmware-Hardware-Abgleich
    if ( deviceUpdate != nullptr )
    {
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Herstellen der Verbindung zur Datenbank ]\n" ) % __FILE__ % __LINE__ % __func__ ).c_str() );

        if ( deviceUpdate->isAvailable() )
        {
            checkBoardRevision = false;
        }
        else
        {
            HardwareInfo hardwareInfo;
            std::string errorMsg;
            checkBoardRevision = !deviceUpdate->open( errorMsg ) || hardwareInfo.isMasterIV( hwInfo );
            if ( !errorMsg.empty() )
            {
                setErrorText( errorMsg );
            }
        }
    }

    // Passende Dateien suchen
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Schleife ueber die %4% Eintraege des Zip-Archives ]\n" ) % __FILE__ % __LINE__ % __func__ % zeZIP.index ).c_str() );

    for (i=0; i < zeZIP.index; i++) {

        if (ZR_OK != GetZipItem(hZip, i, &ze)) {
            // Fehlertext von Zip-API setzen.
            setRecentZipError( __FILE__, __LINE__ );
            trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
            break;
        }

        // Ist kein Ordner?
        if (ze.attr & FILE_ATTRIBUTE_DIRECTORY) {

            continue;
        }
        // Nur ".hex" gueltig.
        size_t pos = strlen(ze.name) - 4;
        if (ze.name[pos] != '.' || 
            !(ze.name[pos + 1] == 'h' || ze.name[pos + 1] == 'H') ||
            !(ze.name[pos + 2] == 'e' || ze.name[pos + 2] == 'E') ||
            !(ze.name[pos + 3] == 'x' || ze.name[pos + 3] == 'X') ) {

                continue;
        }

        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Bearbeitung der Datei %4% Name %5% ]\n" ) % __FILE__ % __LINE__ % __func__ % i % ze.name ).c_str() );

        // Laden der Softwareinfo
        strcpy(tmp, m_iniFile.ReadString(ze.name, "SWI", ""));
        if (!strlen(tmp) || tmp[0] != '#') {
            // Eigenen Fehlertext setzen.
            setErrorText( "Lesen der Softwareinfo(1) fehlgeschlagen." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % errorDetail() ).c_str() );
            goto error_label;
        }

        // Die Softwareinfo in eine Struktur wandeln
        StructSoftwareInfo swInfo = SoftwareInfoStringToStruct(tmp);
        if (!swInfo.size) {
            // Eigenen Fehlertext setzen.
            setErrorText( "Lesen der Softwareinfo(2) fehlgeschlagen." );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
            trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ %__LINE__ %__func__ %errorDetail() ).c_str() );
            goto error_label;
        }

        // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
        iniVersion = m_iniFile.ReadString( ze.name, "Version", "" );
        
        // Die aktuelle Firmwaredatei wird ignoriert, wenn einer der folgenden Fälle eintritt:
        // (1) Version steht auf der Blacklist
        // (2) Version ist älter als die erlaubte Minimalversion (Bei der Ermittlung der Minimalversion wird die Whitelist berücksichtigt.)
        // (3) Es sollte eine minimale Version geben (checkBoardRevision == false), diese ist allerdings undefiniert
        // (4) Die Hardwareinfo passt nicht zur Softwareinfo
        //
        // Es genügt, wenn einer der Fälle eintritt.
        blacklisted = deviceUpdate->isVersionBlacklisted( iniVersion );
        versionToOld = deviceUpdate->minFwVersion() > SoftwarebuildVersion( iniVersion );
        versionUndefined = deviceUpdate->minFwVersion() == SoftwarebuildVersion();
        
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Dateipruefung: iniVersion = %4%, blacklisted = %5%, versionToOld = %6%, versionUndefined = %7%  ]\n" ) % __FILE__ % __LINE__ % __func__ 
                           % iniVersion % blacklisted % versionToOld % versionUndefined ).c_str() );

        if ( blacklisted || versionToOld || (!checkBoardRevision && versionUndefined) ||
             !CompareInfos( hwInfo, swInfo, iniVersion, checkBoardRevision ) )
        {
            errorExtended.append( boost::str( boost::format( "%1%:[%2%]\n" ) % ze.name % _deviceInfo.m_errorText ) );

            trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Resultat Dateipruefung: Ungeeignete Datei ]\n" ) % __FILE__ %__LINE__ %__func__ ).c_str() );

            // Naechste Datei pruefen
            continue;
        }

        // Version halten
        m_version = version;

        // Zeitstempel halten
        m_timestamp = timestamp;

        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Resultat Dateipruefung: Aufnahme in FwVersionListe; version = %4%, timestand = %5%, iniVersion = %6%, ze.name = %7% ]\n" ) % __FILE__ %__LINE__ %__func__ 
                           % version % timestamp % iniVersion % ze.name ).c_str() );

        deviceUpdate->addFwVersion( iniVersion, ze.name );
    }

    if ( !deviceUpdate->latestFwFile().empty() )
    {
        loaded = 1;
    }

error_label:
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Marke error_label ]\n" ) % __FILE__ %__LINE__ %__func__ ).c_str() );

    // Ziparchiv schliessen
    CloseZip(hZip);	
    trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ Schliessen der Inidatei(2) %4% ]\n" ) % __FILE__ %__LINE__ %__func__ % szIniFile ).c_str() );
    // Inidatei schliessen
    m_iniFile.CloseIniFile();
    // Entzippte Dateien loeschen
    if ( !szIniFile.empty() ) ::remove(szIniFile.c_str());
    if ( !szFile.empty() ) ::remove(szFile.c_str());

    // Reset bei fehlerhaftem Laden
    resetData();
    if (errorState() != Error_state::none || !loaded)
    {
        // Es wurde keine passende Datei gefunden.
        setErrorState( Error_state::no_file_match );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ %__LINE__ %errorText() ) );
        trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: %4% ]\n" ) % __FILE__ %__LINE__ %__func__ %errorDetail() ).c_str() );

        // Fehlertext der einzelnen Dateien uebernehmen.
        setErrorTextExtended( errorExtended );

        // Fehlertext uebernehmen
        filePath += ":[" + errorText() + "]";
    }
#endif

    return loaded;
}

int FileDeviceZip::loadFirmwareFileZip( std::string &filePath, const std::string &fwFileName )
{
    if ( fwFileName.empty() || filePath.empty() )
    {
        return 0;
    }

    int loaded = 0;

#if defined(_MSC_VER) || defined(__MINGW32__)
    std::string szFile;
    std::string szIniFile;
    char tmp[256]{};
    int index{};
    ZIPENTRY ze;
    std::unique_ptr<char, void (*)( void * )> md5( nullptr, &free );

    // Fehler zuruecksetzen
    resetError();

    HZIP hZip = OpenZip( filePath.c_str(), 0 );
    if ( hZip == nullptr )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );

        // Fehler anfuegen
        filePath += ":[" + errorText() + "]";
        return 0;
    }

    // Suchen der Initialisierungsdatei der Softwareinfo
    if ( ZR_OK != FindZipItem( hZip, "SWI.ini", false, &index, &ze ) )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto loadFirmwareFileZip_exit;
    }

    // Dateiname zusammenstellen
    szIniFile = Utils::File::uniqueTempFileName( "df_" );
    
    // Entzippen der Initialisierungsdatei
    if ( ZR_OK != UnzipItem( hZip, index, szIniFile.c_str() ) )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto loadFirmwareFileZip_exit;
    }
    
    /*
    // MD5 ueber die Initialisierungsdatei erstellen
    md5.reset( MD5File( szIniFile ) );
    if ( !md5 )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFileZip_exit;
    }
    */

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
    if ( !m_iniFile.OpenIniFile( szIniFile.c_str() ) )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFileZip_exit;
    }

    // Suchen der Firmwaredatei
    if ( ZR_OK != FindZipItem( hZip, fwFileName.c_str(), false, &index, &ze ) )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto loadFirmwareFileZip_exit;
    }

    // Dateiname zusammenstellen
    szFile = Utils::File::uniqueTempFileName( "df_" );
    
    // Entzippen der einzuspielenden Datei
    if ( ZR_OK != UnzipItem( hZip, index, szFile.c_str() ) )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        goto loadFirmwareFileZip_exit;
    }

    // MD5 über Firmwaredatei erstellen
    md5.reset( MD5File( szFile.c_str() ) );
    if ( !md5 ) { 
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) ); 
        goto loadFirmwareFileZip_exit;
    }

    // MD5 der Datei aus Ini lesen
    strcpy(tmp, m_iniFile.ReadString(ze.name, "MD5", ""));

    // Keine Angabe vorhanden. || Datei ist korrupt.
    if ( !strlen( tmp ) || strcmp( tmp, md5.get() ) != 0 )
    {
        // Keine Angabe vorhanden.
        goto loadFirmwareFileZip_exit;
    }

    // Dateidaten laden
    resetData();
    if ( !load( szFile ) )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( boost::str( boost::format( "Laden der Geraetedatei fehlgeschlagen (%1%)." ) % static_cast<int>(errorState()) ) );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFileZip_exit;
    }
    else if ( GetHeaderSoftwareInfo().size )
    {
        // Fehler zuruecksetzen
        resetError();

        // Den Namen der gewaehlten Datei fuer Log anfuegen
        filePath += "\\" + std::string( ze.name );

        // Ist die Softwareinformation in der Geraetedatei verfuegbar?
        loaded = 1;
    }

loadFirmwareFileZip_exit:
    // Ziparchiv schliessen
    CloseZip( hZip );

    // Inidatei schliessen
    m_iniFile.CloseIniFile();

    // Entzippte Dateien loeschen
    if ( !szIniFile.empty() )
    {
        ::remove( szIniFile.c_str() );
    }

    if ( !szFile.empty() )
    {
        ::remove( szFile.c_str() );
    }

    // Reset bei fehlerhaftem laden
    if ( errorState() == Error_state::none && !loaded )
    {
        resetData();

        // Es wurde keine passende Datei gefunden.
        setErrorState( Error_state::no_file_match );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Fehlertext uebernehmen
        filePath += ":[" + errorText() + "]";
    }
#endif

    return loaded;
}

int FileDeviceZip::loadFirmwareFile( std::string &path, const std::string &fwFileName )
{
    if ( fwFileName.empty() || path.empty() || (Utils::File::dir( path ).compare( path ) != 0) )
    {
        return 0;
    }

    int loaded = 0;
    char tmp[256]{};
    std::unique_ptr<char, void (*)( void * )> md5( nullptr, &free );

    std::string iniFilePath = path + "/SWI.ini";
    std::string fwFilePath = path + "/" + fwFileName;

    // Fehler zuruecksetzen
    resetError();

    if ( !Utils::File::exists( iniFilePath ) )
    {
        setErrorText( "Die Datei 'SWI.ini' konnte nicht gefunden werden." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }

    /*
    // MD5 ueber die Initialisierungsdatei erstellen
    md5.reset( MD5File( iniFilePath.c_str() ));
    if ( !md5 )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }
    */

    // Oeffnen der Initialisierungsdatei fuer die Softwareinfo
#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( !m_iniFile.OpenIniFile( iniFilePath.c_str() ) )
#else
    if ( m_iniFile.LoadFile( iniFilePath.c_str() ) != SI_Error::SI_OK )
#endif
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }

    // Suchen der Firmwaredatei
    if ( !Utils::File::exists( fwFilePath ) )
    {
        setErrorText( "Die Firmwaredatei konnte nicht gefunden werden." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }

    // MD5 über Firmwaredatei erstellen
    md5.reset(MD5File( fwFilePath.c_str() ));
    if ( !md5 )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }

    // MD5 der Datei aus Ini lesen
#if defined(_MSC_VER) || defined(__MINGW32__)
    strcpy( tmp, m_iniFile.ReadString( fwFileName.c_str(), "MD5", "" ) );
#else
    strcpy( tmp, m_iniFile.GetValue( fwFileName.c_str(), "MD5", "" ) );
#endif

    // Keine Angabe vorhanden. || Datei ist korrupt.
    if ( !strlen( tmp ) || strcmp( tmp, md5.get() ) != 0 )
    {
        // Keine Angabe vorhanden.
        goto loadFirmwareFile_exit;
    }

    // Dateidaten laden
    resetData();
    if ( !load( fwFilePath.c_str() ) )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( boost::str( boost::format( "Laden der Geraetedatei fehlgeschlagen (%1%)." ) % static_cast<int>(errorState()) ) );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        goto loadFirmwareFile_exit;
    }
    else if ( GetHeaderSoftwareInfo().size )
    {
        // Fehler zuruecksetzen
        resetError();

        // Den Namen der gewaehlten Datei fuer Log anfuegen
        path += "\\" + fwFileName;

        // Ist die Softwareinformation in der Geraetedatei verfuegbar?
        loaded = 1;
    }

loadFirmwareFile_exit:

#if defined(_MSC_VER) || defined(__MINGW32__)
    // Inidatei schliessen
    m_iniFile.CloseIniFile();
#endif

    // Reset bei fehlerhaftem laden
    if ( errorState() == Error_state::none && !loaded )
    {
        resetData();

        // Es wurde keine passende Datei gefunden.
        setErrorState( Error_state::no_file_match );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Fehlertext uebernehmen
        path += ":[" + errorText() + "]";
    }

    return loaded;
}

bool FileDeviceZip::updateDBAvailableZip( const std::string &filePath, std::string &expectedMD5, std::string &errorMsg )
{
#if defined(_MSC_VER) || defined(__MINGW32__)
    // Zurücksetzen des erwarteten MD5 der Datenbank
    expectedMD5.clear();

    if ( filePath.empty() )
    {
        errorMsg = "Fehlende bzw. leere Pfaddefinition.";
        return false;
    }
    
    HZIP hZip = OpenZip( filePath.c_str(), 0 );
    if ( hZip == nullptr )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );

        // Fehler setzen
        errorMsg = errorText();
        return false;
    }

    int index{};
    ZIPENTRY ze;
    std::unique_ptr<char, void (*)( void * )> md5( nullptr, &free );
    std::string iniFileName;

    std::function<bool( bool )> exitFunc = [&]( bool result )
    {
        CloseZip( hZip );
        // Fehler setzen
        errorMsg = errorText();

        m_iniFile.CloseIniFile();
        if ( Utils::File::exists( iniFileName ) )
        {
            ::remove( iniFileName.c_str() );
        }

        return result;
    };

    // Suchen der Initialisierungsdatei der Softwareinfo
    if ( ZR_OK != FindZipItem( hZip, "SWI.ini", false, &index, &ze ) )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return exitFunc( false );
    }

    // MD5 der Ini-Datei (SWI.ini) ermitteln und mit der Initialisierungsdatei der Sicherheit vergleichen,
    // um Manipulationen am Archiv zu erkennen.
    iniFileName = Utils::File::uniqueTempFileName( "df_" );

    // Entpacken der Initialisierungsdatei
    if ( UnzipItem( hZip, index, iniFileName.c_str() ) != ZR_OK )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return exitFunc( false );
    }

    // MD5 über die Initialisierungsdatei erstellen
    md5.reset( MD5File( iniFileName.c_str() ) );
    if ( !md5 )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

    // Dateiname der Sicherheitsinitialisierungsdatei zusammenstellen
    std::string secureIni = boost::str( boost::format( "%1%.ini" ) % md5.get() );
    // Suchen der Initialisierungsdatei der Sicherheit
    if ( FindZipItem( hZip, secureIni.c_str(), false, &index, &ze ) != ZR_OK )
    {
        // Fehlertext von Zip-API setzen.
        setRecentZipError( __FILE__, __LINE__ );
        return exitFunc( false );
    }

    // Öffnen der Initialisierungsdatei für die Softwareinfo
    if ( !m_iniFile.OpenIniFile( iniFileName.c_str() ) )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

    expectedMD5 = m_iniFile.ReadString( "firmwareUpdate.db", "MD5", "" );
    m_iniFile.CloseIniFile();
    return exitFunc( !expectedMD5.empty() );

#else

    return false;

#endif
}

bool FileDeviceZip::updateDBAvailable( const std::string &filePath, std::string &expectedMD5, std::string &errorMsg )
{
    // Zurücksetzen des erwarteten MD5 der Datenbank
    expectedMD5.clear();

    if ( filePath.empty() )
    {
        errorMsg = "Fehlende bzw. leere Pfaddefinition.";
        return false;
    }

    std::unique_ptr<char, void (*)( void * )> md5( nullptr, &free );

    std::function<bool( bool )> exitFunc = [&]( bool result )
    {
        // Fehler setzen
        errorMsg = errorText();

        return result;
    };

    std::string rootPath = Utils::File::dir( filePath );
    // Suchen der Initialisierungsdatei der Softwareinfo
    std::string iniFileName = rootPath + "/SWI.ini";
    if ( !Utils::File::exists( iniFileName ) )
    {
        setErrorText( "Die Datei 'SWI.ini' konnte nicht gefunden werden." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

    // MD5 der Ini-Datei (SWI.ini) ermitteln und mit der Initialisierungsdatei der Sicherheit vergleichen,
    // um Manipulationen am Archiv zu erkennen.
    
    // MD5 über die Initialisierungsdatei erstellen
    md5.reset( MD5File( iniFileName.c_str() ) );
    if ( !md5 )
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Erzeugen des MD5(1) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

    // Dateiname der Sicherheitsinitialisierungsdatei zusammenstellen
    std::string secureIni = boost::str( boost::format( "%1%/%2%.ini" ) % rootPath % md5.get() );
    // Suchen der Initialisierungsdatei der Sicherheit
    if ( !Utils::File::exists( secureIni ) )
    {
        setErrorText( "Die Initialisierungsdatei der Sicherheit konnte nicht gefunden werden." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

    // Öffnen der Initialisierungsdatei für die Softwareinfo
#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( !m_iniFile.OpenIniFile( iniFileName.c_str() ) )
#else
    if ( m_iniFile.LoadFile( iniFileName.c_str() ) != SI_Error::SI_OK )
#endif
    {
        // Eigenen Fehlertext setzen.
        setErrorText( "Oeffnen der Initialisierungsdatei(2) fehlgeschlagen." );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return exitFunc( false );
    }

#if defined(_MSC_VER) || defined(__MINGW32__)
    expectedMD5 = m_iniFile.ReadString( "firmwareUpdate.db", "MD5", "" );
    m_iniFile.CloseIniFile();
#else
    expectedMD5 = m_iniFile.GetValue( "firmwareUpdate.db", "MD5", "" );
#endif

    return exitFunc( !expectedMD5.empty() );
}
