/**
* @file  file_display_grafics.cpp
* @brief Implementierung der Klasse FileDisplayGrafics.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_display_grafics.h"

#include <fstream>

FileDisplayGrafics::FileDisplayGrafics()
{
    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::display_grafic );

    // Initalisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();
}

FileDisplayGrafics::~FileDisplayGrafics()
{
}

void FileDisplayGrafics::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Hilfsstrukturen aufsetzen
    m_pGraficNode = nullptr;
    for ( auto &grafic : m_pGrafic )
    {
        grafic = nullptr;
    }
}

void FileDisplayGrafics::Default( unsigned char count )
{
    // Speicher fuer die Daten.
    m_buffer.resize( sizeof( GRAFICNODE ) + count * sizeof( unsigned int ) + count * sizeof( GRAFIC ) );

    // Hilfsstrukturzeiger
    m_pGraficNode = (GRAFICNODE *)&m_buffer[0];

    // 2.
    // Grafiken init
    // Grundstruktur aufsetzen
    m_pGraficNode->version = 1;
    m_pGraficNode->cnt     = count;
    m_pGraficNode->temp[0] = 0;
    m_pGraficNode->temp[1] = 0;
    m_pGraficNode->ptrGraficArray = sizeof( GRAFICNODE );

    // Zeigerarray und Grafikstrukturzeiger initialisieren
    for ( unsigned char i = 0; i < m_pGraficNode->cnt; ++i )
    {
        // Zeigereintrag auf Grafikstruktur
        *(unsigned int *)&m_buffer[m_pGraficNode->ptrGraficArray + i * sizeof( unsigned int )] = 
            m_pGraficNode->ptrGraficArray + m_pGraficNode->cnt * sizeof( unsigned int ) + i * sizeof( GRAFIC );

        m_pGrafic[i] = (GRAFIC *)&m_buffer[m_pGraficNode->ptrGraficArray + m_pGraficNode->cnt * sizeof( unsigned int ) + i * sizeof( GRAFIC )];
    }

}

void FileDisplayGrafics::setSizeData( std::size_t size )
{
    m_buffer.resize( size );
    // Reinitialisierung der Strukturzeiger.
    // Hilfsstrukturzeiger
    m_pGraficNode = (GRAFICNODE *)&m_buffer[0];
    // Zeigerarray und Grafikstrukturzeiger initialisieren
    for ( int i = 0; i < m_pGraficNode->cnt; ++i )
    {
        m_pGrafic[i] = (GRAFIC *)&m_buffer[m_pGraficNode->ptrGraficArray + m_pGraficNode->cnt * sizeof( unsigned int ) + i * sizeof( GRAFIC )];
    }
}

bool FileDisplayGrafics::load( const std::string &filePath )
{
    if ( FileIntelHex::load( filePath ) )
    {
        // Member durch gelesene Daten initialisieren und pruefen
        if ( m_buffer.size() )
        {
            // Hilfsstrukturzeiger
            m_pGraficNode = (GRAFICNODE *)&m_buffer[0];

            // 2.
            // Grafiken pruefen

            if ( m_pGraficNode->cnt > max_grafic_count )
            {
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
            else
            {
                // Zeigerarray und Grafikstrukturzeiger initialisieren
                for ( unsigned char i = 0; i < m_pGraficNode->cnt; ++i )
                {
                    m_pGrafic[i] = (GRAFIC *)&m_buffer[m_pGraficNode->ptrGraficArray + m_pGraficNode->cnt * sizeof( unsigned int ) + i * sizeof( GRAFIC )];
                }

                // CRC32 pruefen
                unsigned char ucByteCount;
                for ( unsigned char i = 0; i < m_pGraficNode->cnt; ++i )
                {
                    ucByteCount = (unsigned char)((m_pGrafic[i]->width - 1) / 8) + 1;

                    if ( m_pGrafic[i]->crc != crc32( &m_buffer[m_pGrafic[i]->ptrGrafic], ucByteCount * m_pGrafic[i]->height ) )
                    {
                        setErrorState( Error_state::crc_fail );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                }
            }
        }
        else
        {
            // Fehler zuweisen
            setErrorState( Error_state::no_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileDisplayGrafics::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ ) 
{
    bool res = FileIntelHex::save( filePath, appendToFile, appendEndLine );

    if ( res )
    {
        // Die Datei wird fuer einen eventuellen Import in den Firmwarecode als Binaerdatei gespeichert.
        std::string fName( filePath );
        fName.append( ".grafic.bin" );
        std::ofstream binOut( fName, std::ios::binary );

        // Datei als Binär rausschreiben
        if ( binOut.is_open() == true )
        {
            binOut.write( (const char *)m_buffer.data(), m_buffer.size() );
            binOut.close();
        }
    }

    return res;
}

bool FileDisplayGrafics::loadBin( std::ifstream &fs, std::size_t size ) 
{
    // Lesen aus Stream
    bool res = true;

    if ( size == 0 )
    {
        setErrorState( Error_state::no_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else if ( fs.is_open() == false )
    {
        // Stream nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else
    {
        try
        {
            m_buffer.resize( size );
            fs.read( (char *)&m_buffer[0], size );

            // 2.
            // Grafiken pruefen

            // Anzahl der Grafiken setzen
            unsigned char i = m_pGraficNode->cnt; 

            if ( m_pGraficNode->cnt > max_grafic_count )
            {
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                res = false;
            }
            else
            {
                // Zeigerarray und Grafikstrukturzeiger initialisieren
                for ( i = 0; i < m_pGraficNode->cnt; ++i )
                {
                    m_pGrafic[i] = (GRAFIC *)&m_buffer[m_pGraficNode->ptrGraficArray + m_pGraficNode->cnt * sizeof( unsigned int ) + i * sizeof( GRAFIC )];
                }

                // CRC32 pruefen
                unsigned char ucByteCount;
                for ( i = 0; i < m_pGraficNode->cnt; ++i )
                {
                    ucByteCount = (unsigned char)((m_pGrafic[i]->width - 1) / 8) + 1;

                    if ( m_pGrafic[i]->crc != crc32( &m_buffer[m_pGrafic[i]->ptrGrafic], ucByteCount * m_pGrafic[i]->height ) )
                    {
                        setErrorState( Error_state::crc_fail );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        res = false;
                        break;
                    }
                }
            }
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_read );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            res = false;
        }
    }

    return res;
}

bool FileDisplayGrafics::saveBin( std::ofstream &fs ) 
{
    bool res = true;

    if ( fs.is_open() == false )
    {
        // Stream nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else
    {
        try
        {
            fs.write( (const char *)&m_buffer[0], m_buffer.size() );
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_write );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            res = false;
        }
    }

    return res;
}
